/* based on tcpclient from ucspi-tcp-0.88: tcpclient.c
 *
 *	Version 0.92
 *   - code cleanup, adjusted code for usage of actual qlibs
 *   - adjusted option '-4' used together with 'localhost'
 *   - removed obsolete header files
 *  Version 0.91, 20180205, Kai Peter
 *   - switched from 'strerr' to 'errmsg'
 *  Version 0.90, 20171208, Kai Peter
 *   - switched to 'ipv4socket' (removed 'noipv6')
 *   - code cleanup
 *  Version 0.89
 *   - added 'localhost' as valid hostname parameter
 *   - removed remoteinfo
 *   - switched to unistd
*/
#include <unistd.h>
#include <netdb.h>       /* required by 'getservbyname' */
#include "socket.h"
#include "sig.h"
#include "getoptb.h"
#include "fmt.h"
#include "scan.h"
#include "str.h"
#include "ip.h"
#include "fd.h"
#include "stralloc.h"
#include "buffer.h"
#include "errmsg.h"
#include "pathexec.h"
#include "dns.h"

#define WHO "qtcpclient"
#define CONNECT "unable to connect to "

void usage(void) {
  errint(ESOFT,"qtcpclient: usage: qtcpclient \
[ -46hHrRdDqv ] \
[ -i localip ] \
[ -p localport ] \
[ -T timeoutconn ] \
[ -l localname ] \
[ -t timeoutinfo ] \
[ -I interface ] \
host port program");
}

int forcev6 = 0;
int verbosity = 1;
int flagdelay = 1;
unsigned long itimeout = 26;
unsigned long ctimeout[2] = { 0, 0 };
uint32 netif = 0;

char iplocal[16] = { 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0 };
uint16 portlocal = 0;
char *forcelocal = 0;

char ipremote[16];
uint16 portremote;

char *hostname;
static stralloc addresses;
static stralloc moreaddresses;

static stralloc tmp;
static stralloc fqdn;
char strnum[FMT_ULONG];
char ipstr[IP6_FMT];

char seed[128];

int ipv4socket;

int main(int argc,char **argv)
{
  unsigned long u;
  int opt;
  char *x;
  int j;
  int s;
  int cloop;

  dns_random_init(seed);

  close(6);
  close(7);
  sig_ignore(sig_pipe);

  while ((opt = getopt(argc,argv,"46dDvqi:p:t:T:l:I:")) != opteof)
    switch(opt) {
      case '4': ipv4socket = 1; break;
      case '6': forcev6 = 1; break;
      case 'd': flagdelay = 1; break;
      case 'D': flagdelay = 0; break;
      case 'v': loglevel = DEBUG; break;
      case 'q': verbosity = 0; break;
      case 'l': forcelocal = optarg; break;
      case 't': scan_ulong(optarg,&itimeout); break;
      case 'T': j = scan_ulong(optarg,&ctimeout[0]);
        if (optarg[j] == '+') ++j;
        scan_ulong(optarg + j,&ctimeout[1]);
        break;
      case 'i': if (!ip6_scan(optarg,iplocal)) usage(); break;
      case 'I': netif=socket_getifidx(optarg); break;
      case 'p': scan_ulong(optarg,&u); portlocal = u; break;
      default: usage();
    }
  argv += optind;

  if (!verbosity)
    buffer_2->fd = -1;
  hostname = *argv;
  if (!hostname) usage();
  /* possibly ipv4socket is set here through the option "-4" */
  if (!hostname[0] || str_equal(hostname,"0") || str_equal(hostname,"localhost"))
    hostname = (ipv4socket?"127.0.0.1":"::1");

  x = *++argv;
  if (!x) usage();
  if (!x[scan_ulong(x,&u)])
    portremote = u;
  else {
    struct servent *se;
    se = getservbyname(x,"tcp");
    if (!se)
      errint(errno,B("unable to figure out port number for ",x));
    portremote = ntohs(se->s_port);
    /* i continue to be amazed at the stupidity of the s_port interface */
  }

  if (!*++argv) usage();

  if (!stralloc_copys(&tmp,hostname)) errmem;
  if (dns_ip6_qualify(&addresses,&fqdn,&tmp) == -1)
    errint(EHARD,B("temporarily unable to figure out IP address for ",hostname));

  if (addresses.len < 16)
    errint(EHARD,B("no IP address for ",hostname));

  if (addresses.len == 16) {
    ctimeout[0] += ctimeout[1];
    ctimeout[1] = 0;
  }

  for (cloop = 0;cloop < 2;++cloop) {
    if (!stralloc_copys(&moreaddresses,"")) errmem;
    for (j = 0;j + 16 <= addresses.len;j += 4) {
      s = socket_tcp();
      if (s == -1)
        errint(errno,"unable to create socket");
      if (socket_bind(s,iplocal,portlocal,netif) == -1)
        errint(errno,"unable to bind socket");
      if (timeoutconn(s,addresses.s + j,portremote,ctimeout[cloop],netif) == 0)
        goto CONNECTED;
      close(s);
      if (!cloop && ctimeout[1] && (errno == error_timeout)) {
        if (!stralloc_catb(&moreaddresses,addresses.s + j,16)) errmem;
      }
      else {
        strnum[fmt_ulong(strnum,portremote)] = 0;
        if (ip6_isv4mapped(addresses.s+j)) {
          ipstr[ip4_fmt(ipstr,addresses.s + j + 12)] = 0;
        } else {
          ipstr[ip6_fmt(ipstr,addresses.s + j)] = 0; }
        errint(errno,B(CONNECT,ipstr," port ",strnum));
      }
    }
    if (!stralloc_copy(&addresses,&moreaddresses)) errmem;
  }

  _exit(EHARD);

CONNECTED:

  if (!flagdelay)
    socket_tcpnodelay(s); /* if it fails, bummer */

  if(!ipv4socket) ipv4socket = ip6_isv4mapped(iplocal);  /* important */

  if (socket_local(s,iplocal,&portlocal,&netif) == -1)
    errint(errno,"unable to get local address");

  strnum[fmt_ulong(strnum,portlocal)] = 0;
  if (ipv4socket)
    ipstr[ip4_fmt(ipstr,iplocal+12)] = 0;
  else
    ipstr[ip6_fmt(ipstr,iplocal)] = 0;

  x = forcelocal;
  if (!x)
    if (dns_name(&tmp,iplocal) == 0) {
      if (!stralloc_0(&tmp)) errmem;
      x = tmp.s;
    }

  if (socket_remote(s,ipremote,&portremote,&netif) == -1)
    errint(errno,"unable to get remote address");

  if (ipv4socket)
    ipstr[ip4_fmt(ipstr,ipremote+12)] = 0;
  else
    ipstr[ip6_fmt(ipstr,ipremote)] = 0;

  if (loglevel >= INFO)
    log(B("connected to ",ipstr," port ",strnum));

  if (fd_move(6,s) == -1)
    errint(errno,"unable to set up descriptor 6");
  if (fd_copy(7,6) == -1)
    errint(errno,"unable to set up descriptor 7");
  sig_uncatch(sig_pipe);
  pathexec(argv);
  errint(errno,B("unable to run ",*argv));

  return(0);   /* never reached */
}
